# -*- coding: utf-8 -*-
"""
Created on Sat Dec 16 10:16:50 2023
Engine to train the surrogate with

@author: Rebecca Kohlhaas
"""
import h5py
import numpy as np
import os

from .inputs import Input
from .exp_designs import ExpDesigns
from .surrogate_models import MetaModel

class Engine():
    
    
    def __init__(self, MetaMod, Model, ExpDes):
        self.MetaModel = MetaMod
        self.Model = Model
        self.ExpDesign = ExpDes
        
    def start_engine(self) -> None:
        """
        Do all the preparations that need to be run before the actual training

        Returns
        -------
        None

        """
        self.out_names = self.Model.Output.names
        self.MetaModel.out_names = self.out_names
        
        # Add expDesign to MetaModel if it does not have any
        if not hasattr(self.MetaModel, 'ExpDesign'):
            self.MetaModel.ExpDesign = self.ExpDesign
        
    def train_normal(self, parallel = False, verbose = False) -> None:
        """
        Trains surrogate on static samples only.
        Samples are taken from the experimental design and the specified 
        model is run on them.
        Alternatively the samples can be read in from a provided hdf5 file.

        Returns
        -------
        None

        """
            
        ExpDesign = self.ExpDesign
        MetaModel = self.MetaModel
        # Read ExpDesign (training and targets) from the provided hdf5
        if ExpDesign.hdf5_file is not None:
            # TODO: need to run 'generate_ED' as well after this or not?
            ExpDesign.read_from_file(self.out_names)
        else:
            # Check if an old hdf5 file exists: if yes, rename it
            hdf5file = f'ExpDesign_{self.Model.name}.hdf5'
            if os.path.exists(hdf5file):
                os.rename(hdf5file, 'old_'+hdf5file)

        # ---- Prepare X samples ----
        # For training the surrogate use ExpDesign.X_tr, ExpDesign.X is for the model to run on 
        ExpDesign.generate_ED(ExpDesign.n_init_samples,
                                              transform=True,
                                              max_pce_deg=np.max(MetaModel.pce_deg))
        
        # ---- Run simulations at X ----
        if not hasattr(ExpDesign, 'Y') or ExpDesign.Y is None:
            print('\n Now the forward model needs to be run!\n')
            ED_Y, up_ED_X = self.Model.run_model_parallel(ExpDesign.X)
            ExpDesign.Y = ED_Y
        else:
            # Check if a dict has been passed.
            if not type(ExpDesign.Y) is dict:
                raise Exception('Please provide either a dictionary or a hdf5'
                                'file to ExpDesign.hdf5_file argument.')
                
        # Separate output dict and x-values
        if 'x_values' in ExpDesign.Y:
            ExpDesign.x_values = ExpDesign.Y['x_values']
            del ExpDesign.Y['x_values']
        
        MetaModel.ndim = ExpDesign.ndim
        
        # Build and train the MetaModel on the static samples
        MetaModel.CollocationPoints = ExpDesign.X_tr
        MetaModel.build_metamodel()
        
        # Fit the surrogate
        MetaModel.fit(ExpDesign.X_tr, ExpDesign.Y, parallel, verbose)
                
            
    def train_sequential(self, parallel = False, verbose = False) -> None:
        """
        Train the surrogate in a sequential manner.
        First build and train evereything on the static samples, then iterate
        choosing more samples and refitting the surrogate on them.

        Returns
        -------
        None

        """
        self.train_normal(parallel, verbose)