# -*- coding: utf-8 -*-
"""
Test the ExpDesigns class in bayesvalidrox.
Tests are available for the following functions:
    _check_ranges       - x
    fit_dist            - x
    
Class ExpDesigns: 
    check_valid_inputs  - x
    generate_samples
    generate_ED
    read_from_file
    init_param_space    - x
    build_polytypes     - x
    random_sampler
    pcm_sampler
    transform           - x

@author: Rebecca Kohlhaas
"""
import sys
sys.path.append("src/")
import pytest
import numpy as np

from bayesvalidrox.surrogate_models.inputs import Input
import bayesvalidrox.surrogate_models.exp_designs as exp
from bayesvalidrox.surrogate_models.exp_designs import ExpDesigns

#%% Test check_ranges

def test_check_ranges() -> None:
    """
    Check to see if theta lies in expected ranges
    """
    theta = [0.5,1.2]
    ranges = [[0,1],[1,2]]
    assert exp.check_ranges(theta, ranges) == True
    
def test_check_ranges_inv() -> None:
    """
    Check to see if theta lies not in expected ranges
    """
    theta = [1.5,1.2]
    ranges = [[0,1],[1,2]]
    assert exp.check_ranges(theta, ranges) == False
 
#%% Test fit_dist

def test_fit_dist_uniform() -> None:
    """
    Check to see if recognizes uniform dist and parameters sufficiently well.
    """
    y = np.random.uniform(0,1,100000)
    dist, params = exp.fit_dist(y)
    assert dist == 'uniform' and params[0] <0.01 and params[1]>0.99

def test_fit_dist_norm() -> None:
    """
    Check to see if recognizes normal dist and parameters sufficiently well.
    """
    y = np.random.normal(0,1,100000)
    dist, params = exp.fit_dist(y)
    assert dist == 'norm' and params[0] <0.1 and params[1]>0.9

if 0:
    def test_fit_dist_lognorm() -> None:
        """
        Check to see if recognizes lognorm dist and parameters sufficiently well.
        """
        y = np.random.lognormal(0,1,100000)
        dist, params = exp.fit_dist(y)
        assert dist == 'lognorm' and params[0] <0.1 and params[1]>0.9
    
    def test_fit_dist_expon() -> None:
        """
        Check to see if recognizes exponential dist and parameters sufficiently well.
        """
        y = np.random.exponential(0,100000)
        dist, params = exp.fit_dist(y)
        assert dist == 'expon' and params[0] <0.1

#%% Test ExpDesign.check_valid_input

def test_check_valid_input_hasmarg() -> None:
    """
    Distribution not built if no marginals set
    """
    inp = Input()
    with pytest.raises(AssertionError) as excinfo:
        exp = ExpDesigns(inp)
    assert str(excinfo.value) == 'Cannot build distributions if no marginals are given'

def test_check_valid_input_haspriors() -> None:
    """
    Distribution not built if no distribution set for the marginals
    """
    inp = Input()
    inp.add_marginals()
    with pytest.raises(AssertionError) as excinfo:
        exp = ExpDesigns(inp)
    assert str(excinfo.value) ==  'Not all marginals were provided priors'
    
def test_check_valid_input_priorsmatch() -> None:
    """
    Distribution not built if dist types do not align
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    inp.add_marginals()
    inp.Marginals[1].dist_type = 'normal'
    inp.Marginals[1].parameters = [0,1]
    with pytest.raises(AssertionError) as excinfo:
        exp = ExpDesigns(inp)
    assert str(excinfo.value) == 'Distributions cannot be built as the priors have different types'

def test_check_valid_input_samples() -> None:
    """
    Design built correctly - samples
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    inp.add_marginals()
    inp.Marginals[1].input_data = x+2
    try:
        exp = ExpDesigns(inp)
    except AssertionError:
        pytest.fail("ExpDesign raised AssertionError unexpectedly!")
    # TODO: check for better options to assert that no error at all occurred
    

def test_check_valid_input_both() -> None:
    """
    Design no built - samples and dist type given
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    inp.Marginals[0].dist_type = 'normal'
    with pytest.raises(AssertionError) as excinfo:
        exp = ExpDesigns(inp)
    assert str(excinfo.value) == 'Both samples and distribution type are given. Please choose only one.'

#def test_check_valid_input_distnotok() -> None:
#    """
#    Design built incorrectly - dist types without parameters
#    """
#    inp = Input()
#    inp.add_marginals()
#    inp.Marginals[0].dist_type = 'normal'
#    inp.add_marginals()
#    inp.Marginals[1].dist_type = 'normal'
#    with pytest.raises(AssertionError) as excinfo:
#        exp = ExpDesigns(inp)
#    assert str(excinfo.value) == 'Some distributions do not have characteristic values'
    
def test_check_valid_input_distok() -> None:
    """
    Design built correctly - dist types
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0,1]
    inp.add_marginals()
    inp.Marginals[1].dist_type = 'normal'
    inp.Marginals[1].parameters = [0,1]
    try:
        exp = ExpDesigns(inp)
    except AssertionError:
        pytest.fail("ExpDesign raised AssertionError unexpectedly!")
    # TODO: check for better options to assert that no error at all occurred

#%% Test ExpDesign.build_polytypes
def test_build_polytypes_normalerr() -> None:
    """
    Build dist 'normal' - too few params
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = []
    exp = ExpDesigns(inp)
    with pytest.raises(AssertionError) as excinfo:
        exp.build_polytypes(False)
    assert str(excinfo.value) == 'Distribution has too few parameters!'

def test_build_polytypes_normal() -> None:
    """
    Build dist 'normal'
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0,1]
    exp = ExpDesigns(inp)
    exp.build_polytypes(False)
    
def test_build_polytypes_uniferr() -> None:
    """
    Build dist 'unif' - too few params
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'unif'
    inp.Marginals[0].parameters = []
    exp = ExpDesigns(inp)
    with pytest.raises(AssertionError) as excinfo:
        exp.build_polytypes(False)
    assert str(excinfo.value) == 'Distribution has too few parameters!'

def test_build_polytypes_unif() -> None:
    """
    Build dist 'unif'
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'unif'
    inp.Marginals[0].parameters = [0,1]
    exp = ExpDesigns(inp)
    exp.build_polytypes(False)
    
def test_build_polytypes_gammaerr() -> None:
    """
    Build dist 'gamma' - too few params
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'gamma'
    inp.Marginals[0].parameters = []
    exp = ExpDesigns(inp)
    with pytest.raises(AssertionError) as excinfo:
        exp.build_polytypes(False)
    assert str(excinfo.value) == 'Distribution has too few parameters!'

def test_build_polytypes_gamma() -> None:
    """
    Build dist 'gamma'
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'gamma'
    inp.Marginals[0].parameters = [0,1,0]
    exp = ExpDesigns(inp)
    with pytest.raises(ValueError) as excinfo:
        exp.build_polytypes(False)
    assert str(excinfo.value) == 'Parameter values are not valid, please set differently'
    
def test_build_polytypes_betaerr() -> None:
    """
    Build dist 'beta' - too few params
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'beta'
    inp.Marginals[0].parameters = []
    exp = ExpDesigns(inp)
    with pytest.raises(AssertionError) as excinfo:
        exp.build_polytypes(False)
    assert str(excinfo.value) == 'Distribution has too few parameters!'

def test_build_polytypes_beta() -> None:
    """
    Build dist 'beta'
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'beta'
    inp.Marginals[0].parameters = [0.5,1,2,3]
    exp = ExpDesigns(inp)
    exp.build_polytypes(False)
    
        
def test_build_polytypes_lognormerr() -> None:
    """
    Build dist 'lognorm' - too few params
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'lognorm'
    inp.Marginals[0].parameters = []
    exp = ExpDesigns(inp)
    with pytest.raises(AssertionError) as excinfo:
        exp.build_polytypes(False)
    assert str(excinfo.value) == 'Distribution has too few parameters!'

def test_build_polytypes_lognorm() -> None:
    """
    Build dist 'lognorm'
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'lognorm'
    inp.Marginals[0].parameters = [0.5,1,2,3]
    exp = ExpDesigns(inp)
    exp.build_polytypes(False)
    
        
def test_build_polytypes_exponerr() -> None:
    """
    Build dist 'expon' - too few params
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'beta'
    inp.Marginals[0].parameters = []
    exp = ExpDesigns(inp)
    with pytest.raises(AssertionError) as excinfo:
        exp.build_polytypes(False)
    assert str(excinfo.value) == 'Distribution has too few parameters!'

def test_build_polytypes_expon() -> None:
    """
    Build dist 'expon'
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'expon'
    inp.Marginals[0].parameters = [0.5,1,2,3]
    exp = ExpDesigns(inp)
    exp.build_polytypes(False)
    
        
def test_build_polytypes_weibullerr() -> None:
    """
    Build dist 'weibull' - too few params
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'weibull'
    inp.Marginals[0].parameters = []
    exp = ExpDesigns(inp)
    with pytest.raises(AssertionError) as excinfo:
        exp.build_polytypes(False)
    assert str(excinfo.value) == 'Distribution has too few parameters!'

def test_build_polytypes_weibull() -> None:
    """
    Build dist 'weibull'
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'weibull'
    inp.Marginals[0].parameters = [0.5,1,2,3]
    exp = ExpDesigns(inp)
    exp.build_polytypes(False)
    

def test_build_polytypes_arbitrary() -> None:
    """
    Build poly 'arbitrary'
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.build_polytypes(False)
   
def test_build_polytypes_rosenblatt() -> None:
    """
    Build dist with rosenblatt
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.build_polytypes(True)
    
def test_build_polytypes_samples() -> None:
    """
    Build dist from samples
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.build_polytypes(False)
    
    
def test_build_polytypes_samples2d() -> None:
    """
    Build dist from samples - samples too high dim
    """
    x = np.random.uniform(0,1,(2,1000))
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    with pytest.raises(ValueError) as excinfo:
        exp.build_polytypes(False)
    assert str(excinfo.value) == 'The samples provided to the Marginals should be 1D only'
    
#%% Test ExpDesign.init_param_space

def test_init_param_space_nomaxdegsample() -> None:
    """
    Init param space without max_deg for given samples
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.init_param_space()

def test_init_param_space_nomaxdegdist() -> None:
    """
    Init param space without max_deg for given dist
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'expon'
    inp.Marginals[0].parameters = [0.5,1,2,3]
    exp = ExpDesigns(inp)
    exp.init_param_space()
     
def test_init_param_space_maxdeg() -> None:
    """
    Init param space with max_deg for given samples
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.init_param_space(max_deg=2)
    
def test_init_param_space_maxdegdist() -> None:
    """
    Init param space with max_deg for given dist
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'expon'
    inp.Marginals[0].parameters = [0.5,1,2,3]
    exp = ExpDesigns(inp)
    exp.init_param_space(max_deg=2)
     
    
#%% Test ExpDesign.transform
    
def test_transform_noparamspace() -> None:
    """
    Call transform without a built JDist
    """
    x = np.random.uniform(0,1,1000)
    y = np.random.uniform(0,1,(2,1000))
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    with pytest.raises(AttributeError) as excinfo:
        exp.transform(y)
    assert str(excinfo.value) == 'Call function init_param_space first to create JDist'
      
def test_transform_dimerrlow() -> None:
    """
    Call transform with too few dimensions
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.init_param_space(max_deg=2)
    with pytest.raises(AttributeError) as excinfo:
        exp.transform(x)
    assert str(excinfo.value) == 'X should have two dimensions'
          
def test_transform_dimerrhigh() -> None:
    """
    Call transform with too many dimensions
    """
    x = np.random.uniform(0,1,1000)
    y = np.random.uniform(0,1,(1,1,1000))
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.init_param_space(max_deg=2)
    with pytest.raises(AttributeError) as excinfo:
        exp.transform(y)
    assert str(excinfo.value) == 'X should have two dimensions'
    
def test_transform_dimerr0() -> None:
    """
    Call transform with wrong X.shape[0]
    """
    x = np.random.uniform(0,1,1000)
    y = np.random.uniform(0,1,(2,1000))
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.init_param_space(max_deg=2)
    with pytest.raises(AttributeError) as excinfo:
        exp.transform(y)
    assert str(excinfo.value) == 'The second dimension of X should be the same size as the number of marginals in the InputObj'
   
def test_transform_paramspace() -> None:
    """
    Transform successfully
    """
    x = np.random.uniform(0,1,1000)
    y = np.random.uniform(0,1,(1000,1))
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.init_param_space(max_deg=2)
    exp.transform(y)
    
#%% Test ExpDesign.pcm_sampler

def test_pcm_sampler_noinit() -> None:
    """
    Sample via pcm without init_param_space
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.pcm_sampler(4,2)
    
def test_pcm_sampler_lowdeg() -> None:
    """
    Sample via pcm with init_param_space and small max_deg
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.init_param_space()
    exp.pcm_sampler(4,2)
    
def test_pcm_sampler_highdeg() -> None:
    """
    Sample via pcm with init_param_space and high max_deg
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.init_param_space()
    exp.pcm_sampler(4,4)
    
    
#%% Test ExpDesign.random_sampler

def test_random_sampler() -> None:
    """
    Sample randomly, init_param_space implicitly
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    exp.random_sampler(4)
    
#%% Test ExpDesign.generate_samples

def test_generate_samples() -> None:
    """
    Generate samples according to chosen scheme
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    samples = exp.generate_samples(4)


#%% Test ExpDesign.generate_ED

def test_generate_ED() -> None:
    """
    Generate ED as is
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp)
    samples = exp.generate_ED(4)
    
def test_generate_ED_negsamplenum():
    """
    Generate ED for neg number of samples
    """
    x = np.random.uniform(0,1,1000)
    X = np.random.uniform(0,1,(1,10))
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp, sampling_method = 'user')   
    with pytest.raises(ValueError) as excinfo:
        exp.generate_ED(-1)
    assert str(excinfo.value) == 'A negative number of samples cannot be created. Please provide positive n_samples'
    
    

def test_generate_ED_usernoX() -> None:
    """
    User-defined ED without samples
    """
    x = np.random.uniform(0,1,1000)
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp, sampling_method = 'user')
    with pytest.raises(AttributeError) as excinfo:
        samples = exp.generate_ED(4)
    assert str(excinfo.value) == 'User-defined sampling cannot proceed as no samples provided. Please add them to this class as attribute X'

def test_generate_ED_userXdimerr() -> None:
    """
    User-defined ED with wrong shape of samples
    """
    x = np.random.uniform(0,1,1000)
    X = np.random.uniform(0,1,(2,1,1000))
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp, sampling_method = 'user')
    exp.X = X
    with pytest.raises(AttributeError) as excinfo:
        samples = exp.generate_ED(4)
    assert str(excinfo.value) == 'The provided samples shuld have 2 dimensions'
    

def test_generate_ED_userX() -> None:
    """
    User-defined ED with wrong shape of samples
    """
    x = np.random.uniform(0,1,1000)
    X = np.random.uniform(0,1,(3,1000))
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp, sampling_method = 'user')
    exp.X = X
    samples = exp.generate_ED(4)
    
    
#%% Test ExpDesign.read_from_file

def test_read_from_file():
    """
    No file given to read in
    """
    x = np.random.uniform(0,1,1000)
    X = np.random.uniform(0,1,(1,10))
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].input_data = x
    exp = ExpDesigns(inp, sampling_method = 'user')
    with pytest.raises(AttributeError) as excinfo:
        exp.read_from_file(['Out'])
    assert str(excinfo.value) == 'ExpDesign cannot be read in, please provide hdf5 file first'
    
# TODO: test with provided file


#%% Test ExpDesign.generate_ED(self, n_samples, transform=False, max_pce_deg=None)




if __name__ == '__main__':
    None
