# -*- coding: utf-8 -*-
"""
Test the MetaModel class in bayesvalidrox.
Tests are available for the following functions
Class MetaModel: 
    build_metamodel
    create_metamodel *not used again
    train_norm_design 
    update_pce_coeffs
    create_basis_indices --removed, just redirects
    add_ExpDesign                                   -x
    generate_ExpDesign
    univ_basis_vals
    create_psi
    fit
    adaptive_regression
    corr_loocv_error
    pca_transformation
    gaussian_process_emulator
    eval_metamodel
    create_model_error
    eval_model_error
    auto_vivification
    copy_meta_model_opts
    __select_degree
    generate_polynomials

@author: Rebecca Kohlhaas
"""
import sys
sys.path.append("src/")
import pytest

from bayesvalidrox.surrogate_models.inputs import Input
from bayesvalidrox.surrogate_models.exp_designs import ExpDesigns
from bayesvalidrox.surrogate_models.surrogate_models import MetaModel
from bayesvalidrox.pylink.pylink import PyLinkForwardModel as PL

#%% Test MetaMod constructor on its own

def test_metamod() -> None:
    """
    Construct MetaModel without inputs
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0,1]
    mod = PL()
    mm = MetaModel(inp, mod)
    
#%% Test MetaModel.build_metamodel

def test_build_metamodel_nosamples() -> None:
    """
    Build MetaModel without samples
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0,1]
    mod = PL()
    mm = MetaModel(inp, mod)
    with pytest.raises(AttributeError) as excinfo:
        mm.build_metamodel()
    assert str(excinfo.value) == 'Please provide samples to the metamodle before building it.'
    

def test_build_metamodel() -> None:
    """
    Build MetaModel without inputs
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0,1]
    mod = PL()
    mm = MetaModel(inp, mod)
    mm.build_metamodel()
    'Please provide samples to the metamodle before building it.'


#%% Test MetaMod.generate_polynomials

def test_generate_polynomials_noExp() -> None:
    """
    Generate polynomials without ExpDeg
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0,1]
    mod = PL()
    mm = MetaModel(inp, mod)
    with pytest.raises(AttributeError) as excinfo:
        mm.generate_polynomials()
    assert str(excinfo.value) == 'Generate or add ExpDesign before generating polynomials'
    
def test_generate_polynomials_nodeg() -> None:
    """
    Generate polynomials without max_deg
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0,1]
    mod = PL()
    mm = MetaModel(inp, mod)
    with pytest.raises(AttributeError) as excinfo:
        mm.generate_polynomials()
    assert str(excinfo.value) == ''
    

def test_generate_polynomials_deg() -> None:
    """
    Generate polynomials with max_deg
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0,1]
    mod = PL()
    mm = MetaModel(inp, mod)
    mm.generate_polynomials(4)
    
    
#%% Test MetaMod.add_ExpDesign

def test_add_ExpDesign() -> None:
    """
    Add ExpDesign in MetaModel
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0,1]
    mod = PL()
    mm = MetaModel(inp, mod)
    mm.add_ExpDesign()
    
#%% Test MetaModel.generate_ExpDesign

def test_generate_ExpDesign() -> None:
    """
    Generate ExpDesign in MetaModel
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0,1]
    mod = PL()
    mm = MetaModel(inp, mod)
    mm.generate_ExpDesign(mod, 10)
    # TODO: model not working for this, create a workaround/test model!
    

    
    
if __name__ == '__main__':
    None
